local SKIN = "jetmetalsonic"

//Set default values
local MAX_CHARGES = 2								--Maximum amount of charges per jump
local SUPER_MAX_CHARGES = 4							--Maximum amount of charges per jump when super

rawset(_G, "FULL_CHARGE_BUILDUP", 17*TICRATE/15)	--Amount of charge at maximum, set in global table for HUD script
local HALF_CHARGE_BUILDUP = FULL_CHARGE_BUILDUP/2	--Amount of charge at half-full
local MAX_CHARGE_TIME = 2*TICRATE					--Maximum amount of time allowed to be charging

local BASE_BUILDUP_MOD = 5*FRACUNIT/4				--Buildup modifier at minimum
local ADD_BUILDUP_MOD = 6*FRACUNIT/5				--Buildup modifier at full charge

local BASE_CHRGJMP_HEIGHT = 15*FRACUNIT/2			--Base value for Charge Jump height
local BASE_CHRGTHOK_SPEED = 20*FRACUNIT				--Base value for Charge Thok speed
local WALLBOUNCE_WINDOW = 4*TICRATE/5				--Window of opportunity to perform the wall bounce
local BASE_V_STMPSPEED = -20*FRACUNIT				--Base value for vertical Stomp speed
local BASE_H_STMPSPEED = 6*FRACUNIT					--Base value for horizontal Stomp speed
local BASE_GBOUNCE_HEIGHT = 17*FRACUNIT				--Base value for Ground Bounce height

//Function to reset charging variables
rawset(_G, "ResetCharge", function(p)
	p.chrgbuildUp = 0
	p.charging = false
	p.chargepercent = 0
	p.buildupmod = 0
	p.effsize = 0
	p.halfchargeeff = false
	p.fullchargeeff = false
end)

//Function to reset charge ability variables
rawset(_G, "ResetAbilities", function(p)
	p.chrgjumpUsed = false
	p.superchrgjumpUsed = false
	if p.chrgStomping
		p.chrgStomping = false
		p.prevChrgStomp = true
	end
	p.prepchrgBounce = false
	p.chrgWBounce = 0
	p.chrgGBounce = false
end)

//Function to convert charge into a usable variable and set effect size
local function ConvertCharge(p)
	p.effsize = 3*FRACUNIT/5
	if p.chrgbuildUp > FULL_CHARGE_BUILDUP
		p.chrgbuildUp = FULL_CHARGE_BUILDUP
		p.effsize = 9*FRACUNIT/10
	elseif p.chrgbuildUp > HALF_CHARGE_BUILDUP
		p.effsize = 4*FRACUNIT/5
	end
	p.buildupmod = BASE_BUILDUP_MOD+FixedMul(ADD_BUILDUP_MOD, FixedDiv(p.chrgbuildUp*FRACUNIT, FULL_CHARGE_BUILDUP*FRACUNIT))
end

//Function to break breakable floors, adapted from MAXControl by Inferno
local function FloorBreak(mo, sec)
	for x = mo.x, mo.x + mo.radius, mo.radius
		for y = mo.y, mo.y + mo.radius, mo.radius
			for fof in sec.ffloors()
				if (fof.flags & FF_EXISTS) and (fof.flags & FF_BUSTUP)
				and (mo.z + mo.height == fof.bottomheight or mo.z == fof.topheight
				or (mo.eflags & MFE_VERTICALFLIP) and (mo.z - mo.height == fof.topheight or mo.z == fof.bottomheight))
					EV_CrumbleChain(fof)
				end
			end
		end
	end
end

local function SuperNuke(p, radius)
	if p.powers[pw_super]
		P_StartQuake(20*FRACUNIT, 4)
		P_NukeEnemies(p.mo, p.mo, radius)
		S_StartSound(p.mo, sfx_bkpoof)
	end
end

local function SpawnChargeEffect(p, color)
	local chrgeff = P_SpawnMobj(p.mo.x, p.mo.y, p.mo.z, MT_THOK)
	chrgeff.state = S_CHRGEFF1
	chrgeff.color = color
	chrgeff.scale = p.mo.scale/2
	chrgeff.tracer = p.mo
	chrgeff.ischrgeff = true
end

local function SpawnSparkleEffect(p, color)
	local mo = p.mo
	local sparkle = P_SpawnMobj(mo.x + P_RandomRange(-20, 20) * mo.scale, mo.y + P_RandomRange(-20, 20) * mo.scale, mo.z + P_RandomRange(-8, 56) * mo.scale, MT_THOK)
	sparkle.state = S_CHRGSPARKLE
	if sparkle.z > mo.z + 24 * mo.scale
		P_SetObjectMomZ(sparkle, 2*FRACUNIT, false)
	else
		P_SetObjectMomZ(sparkle, -2*FRACUNIT, false)
	end
	P_InstaThrust(sparkle, R_PointToAngle2(mo.x, mo.y, sparkle.x, sparkle.y), FixedMul(2*FRACUNIT, mo.scale))
	sparkle.colorized = true
	sparkle.color = color
	sparkle.scale = mo.scale
end

//Hook charging to double jumping
addHook("AbilitySpecial", function(p)
	if not (p.mo and p.mo.health and p.mo.valid and p.mo.skin == SKIN) return end
	if not p.charging
	and not (p.pflags & (PF_SPINDOWN|PF_JUMPDOWN))
		if p.charges > 0

			--Start Charging
			S_StartSound(p.mo, sfx_cdfm44)
			p.chrgWBounce = 0
			p.chrgGBounce = false
			p.charging = true
			if p.chrgStomping
				p.chrgStomping = false
				p.prevChrgStomp = true
			end
			p.storeSpd = FixedHypot(p.rmomx, p.rmomy)
			p.pflags = ($|PF_NOJUMPDAMAGE) & ~PF_SPINNING
			p.ghostTrail = 0
		else
			
		end
	end
	return true
end)

//Hook Charging behaviour and follow-up buttons in the PreThinkFrame
addHook("PreThinkFrame", do
	for p in players.iterate
		local mo = p.mo
		if not (mo and mo.health and mo.valid and mo.skin == SKIN) return end
		if p.charges == nil
			p.charges = 0
		end

		--Set visual effect colour
		if p.menu.contents[1][3].value == 0
			p.effcolor = SKINCOLOR_YELLOW
		elseif p.menu.contents[1][3].value == 1
			p.effcolor = SKINCOLOR_ORANGE
		elseif p.menu.contents[1][3].value == 2
			p.effcolor = SKINCOLOR_CRIMSON
		elseif p.menu.contents[1][3].value == 3
			p.effcolor = SKINCOLOR_AETHER
		end

		if p.charging
		and not P_IsObjectOnGround(mo)

			--Lose charging state and all charges when hurt
			if P_PlayerInPain(p)
				ResetCharge(p)
				p.charges = 0
			end

			--Lose charging state when hitting a spring
			if mo.eflags & MFE_SPRUNG
				ResetCharge(p)
			end

			--Set charging animation
			mo.state = S_PLAY_WALK
			mo.sprite2 = SPR2_CHRG
			mo.frame = A

			--Flash colours if not super
			if not p.powers[pw_super]
				if not p.prevcolor
					p.prevcolor = mo.color
				end
				if leveltime % 4 == 0
					mo.color = p.effcolor
				elseif leveltime % 4 == 2
					mo.color = p.prevcolor
				end
			end
			if leveltime % 4 == 0
				mo.colorized = true
			elseif leveltime % 4 == 2
				mo.colorized = false
			end

			--If the JUMP button is held then increase charge
			if (p.cmd.buttons & BT_JUMP)
			and p.chrgbuildUp < MAX_CHARGE_TIME
				p.chrgbuildUp = $+1

				--Charge 50 percent faster when super
				if p.powers[pw_super]
				and leveltime % 2 == 0
				and p.chrgbuildUp < MAX_CHARGE_TIME
					p.chrgbuildUp = $+1
				end

				--Reduce momentum and fall slowly while charging
				mo.momx = 8*$/9
				mo.momy = 8*$/9
				mo.momz = 2*$/5

				--Audio and visual cues for charge levels
				if p.chrgbuildUp >= HALF_CHARGE_BUILDUP
				and not p.halfchargeeff
					S_StartSound(mo, sfx_s3ka8)
					p.chrgSparkle = true
					SpawnChargeEffect(p, p.effcolor)
					p.halfchargeeff = true
				elseif p.chrgbuildUp >= FULL_CHARGE_BUILDUP
				and not p.fullchargeeff
					S_StartSound(mo, sfx_s3ka8)
					SpawnChargeEffect(p, p.effcolor)
					p.fullchargeeff = true
					p.prepchrgBounce = true --Bounce possible
				end

				--If the SPIN button is pressed while charging, perform the Charge Jump
				if (p.cmd.buttons & BT_SPIN)
				and not (p.pflags & PF_SPINDOWN)
					p.pflags = $|PF_SPINDOWN
					if (not p.chrgjumpUsed or (p.powers[pw_super] and not p.superchrgjumpUsed))
					and p.chrgbuildUp < MAX_CHARGE_TIME
						ConvertCharge(p)
						SpawnLaunchEffect(p, 0, p.effsize, p.effcolor)

						local JUMPHEIGHT = FixedMul(p.buildupmod, BASE_CHRGJMP_HEIGHT) --Calculate Jump Height
						if p.powers[pw_super]
							JUMPHEIGHT = 6*$/5 --Increase jump height by 20% when Super
						end
						if mo.eflags & MFE_UNDERWATER
							JUMPHEIGHT = 3*$/5 --Reduce jump height by 40% when underwater
						end
						P_SetObjectMomZ(mo, JUMPHEIGHT, false)
						P_InstaThrust(mo, mo.angle, FixedMul(p.storeSpd/4, mo.scale))  --Give forward speed equal to 25% stored speed
						S_StartSound(mo, sfx_beflap)
						mo.state = S_PLAY_CHRGJMP
						p.ghostTrail = 14

						if not p.prepchrgBounce
							p.prepchrgBounce = false
						end
						if p.prepchrgBounce
							p.prepchrgBounce = false
						end

						SuperNuke(p, 475*FRACUNIT)

						--Lose a charge
						p.charges = $-1
						if p.chrgjumpUsed
							p.superchrgjumpUsed = true
						end
						p.chrgjumpUsed = true
						ResetCharge(p)
					end
				end

			--If the JUMP button is released while charging, perform the Charge Thok or Charge Stomp
			elseif not (p.cmd.buttons & BT_JUMP)
			and p.chrgbuildUp < MAX_CHARGE_TIME
				S_StartSound(mo, sfx_beflap)

				--If holding any direction then perform Charge Thok
				if P_GetPlayerControlDirection(p) ~= 0
					ConvertCharge(p)
					SpawnLaunchEffect(p, 1, p.effsize, p.effcolor)

					local THOKSPEED = (2*p.storeSpd/3)+FixedMul(p.buildupmod, BASE_CHRGTHOK_SPEED) --Calculate Thok speed
					if THOKSPEED > p.thokspeedcap
						THOKSPEED = p.thokspeedcap
					end
					if p.powers[pw_super]
						THOKSPEED = 6*$/5 --Increase thok speed by 20% when Super
					end
					if mo.eflags & MFE_UNDERWATER
						THOKSPEED = 3*$/5  --Reduce thok speed by 40% when underwater
					end
					P_InstaThrust(mo, mo.angle, FixedMul(THOKSPEED, mo.scale))
					if mo.eflags & MFE_UNDERWATER
						P_SetObjectMomZ(mo, 3*FRACUNIT, false)
					else
						P_SetObjectMomZ(mo, 5*FRACUNIT, false)
					end
					mo.state = S_PLAY_CHRGTHK

					if not p.prepchrgBounce
						p.prepchrgBounce = false
					end
					if p.prepchrgBounce
						p.chrgWBounce = WALLBOUNCE_WINDOW
						p.ghostTrail = WALLBOUNCE_WINDOW
						p.prepchrgBounce = false
					else
						p.ghostTrail = 14
					end

					SuperNuke(p, 475*FRACUNIT)
					p.pflags = $ & ~PF_NOJUMPDAMAGE

					--Lose a charge
					p.charges = $-1
					ResetCharge(p)

				--Otherwise, perform Charge Stomp
				else
					ConvertCharge(p)
					SpawnLaunchEffect(p, 0, p.effsize, p.effcolor)

					p.STOMPSPEED = FixedMul(p.buildupmod, BASE_V_STMPSPEED)
					if p.powers[pw_super]
						p.STOMPSPEED = 6*$/5 --Increase stomp speed by 20% when Super
					end
					if mo.eflags & MFE_UNDERWATER
						p.STOMPSPEED = 3*$/5  --Reduce stomp speed by 40% when underwater
					end
					mo.momx = 0
					mo.momy = 0
					if p.effsize > 0
						P_InstaThrust(mo, mo.angle, FixedMul(p.storeSpd/2+BASE_H_STMPSPEED, mo.scale))
					end
					P_SetObjectMomZ(mo, p.STOMPSPEED, false)
					mo.state = S_PLAY_ROLL
					p.ghostTrail = 14

					if not p.prepchrgBounce
						p.prepchrgBounce = false
					end
					if p.prepchrgBounce
						p.chrgGBounce = true
						p.chrgGBouncing = true
						p.prepchrgBounce = false
					end

					SuperNuke(p, 475*FRACUNIT)
					p.pflags = $ & ~PF_NOJUMPDAMAGE

					--Lose a charge
					p.charges = $-1
					p.chrgStomping = true
					ResetCharge(p)
				end

			--If charging longer than MAX_CHARGE_TIME cancel charge and lose all charges
			else
				mo.state = S_PLAY_FALL
				ResetCharge(p)
				p.pflags = $|PF_NOJUMPDAMAGE|PF_THOKKED
				p.charges = 0
			end
		else
			S_StopSoundByID(mo, sfx_cdfm44)
			if p.prevcolor
				mo.color = p.prevcolor
				p.prevcolor = nil
			end
			mo.colorized = false
		end
	end
end)

//Visual effects hooked into ThinkFrames
addHook("ThinkFrame", do
	for p in players.iterate
		local mo = p.mo
		if mo and mo.valid and mo.skin == SKIN

			--Sparkle while sparkles are turned on
			if p.chrgSparkle
				if p.chrgbuildUp >= HALF_CHARGE_BUILDUP
				and leveltime % 2 == 0
					SpawnSparkleEffect(p, p.effcolor)

				--Greatly increase sparkle intensity with charge build-up
				elseif p.chrgbuildUp >= FULL_CHARGE_BUILDUP
				and leveltime % 2 == 1
					SpawnSparkleEffect(p, p.effcolor)
					SpawnSparkleEffect(p, p.effcolor)
					SpawnSparkleEffect(p, p.effcolor)
				end
			end

			--Reset sparkles when charging stops
			if not p.charging
				if p.chrgSparkle
					p.chrgSparkle = false
				end
			end
		end
	end
end)

addHook("PlayerThink", function(p)
	local mo = p.mo
	if not (mo and mo.valid and mo.health and mo.skin == SKIN) return end

	if (maptol & TOL_NIGHTS) return end
	if p.powers[pw_carry] return end

	p.thokspeedcap = 3*skins[mo.skin].normalspeed

	--Return jump damage when entering a rolling state in the air (used for wall bounce)
	if mo.state == S_PLAY_ROLL
	and not P_IsObjectOnGround(mo)
		p.pflags = $ & ~PF_NOJUMPDAMAGE
	end

	--Return to rolling state at the height of a Charge Jump
	if mo.state == S_PLAY_CHRGJMP
		if (mo.momz < 5*FRACUNIT and not (mo.eflags & MFE_VERTICALFLIP))
		or (mo.momz > -5*FRACUNIT and mo.eflags & MFE_VERTICALFLIP)
			mo.state = S_PLAY_ROLL
		end
	end

	--Disable bounce behaviour when hitting a spring or entering goo
	if p.eflags == MFE_SPRUNG
	or mo.state == S_PLAY_SPRING
	or mo.eflags & MFE_GOOWATER
		if p.chrgWBounce
			p.chrgWBounce = 0
		end
		if p.chrgGBounce
			p.chrgGBounce = false
			if p.chrgStomping
				p.chrgStomping = false
				p.prevChrgStomp = true
			end

			--Increase Z-momentum by 75% when hitting a spring with a stomp
			if p.eflags == MFE_SPRUNG
			or mo.state == S_PLAY_SPRING
				P_SetObjectMomZ(mo, 3*mo.momz/4, true)
			end
		end
	end

	//This next bit is to make comboing and ground bouncing play nice with each other
	if not p.chrgGBouncing
		p.chrgGBouncing = false
	end
	if p.chrgGBouncing
	and not p.chrgGBounce
		p.chrgGBouncing = false
	end

	--Reset everything air-related when on the ground and not bouncing
	if (P_IsObjectOnGround(mo) or (mo.eflags & MFE_JUSTHITFLOOR))
		if p.chrgStomping
			p.chrgStomping = false
			p.prevChrgStomp = true
		end
		if not p.chrgGBounce
			p.chrgGBounce = false
			ResetCharge(p)
			ResetAbilities(p)
			if not p.powers[pw_super]
				p.charges = MAX_CHARGES
			else
				p.charges = SUPER_MAX_CHARGES
			end
			p.ghostTrail = 0
			p.pflags = $ & ~PF_NOJUMPDAMAGE

		--Ground bounce behaviour
		else
			p.chrgGBounce = false
			mo.state = S_PLAY_JUMP
			p.pflags = ($ & ~PF_SPINNING)|PF_JUMPED
			FloorBreak(mo, mo.subsector.sector)

			-- Ground bounce dust from Max Control Pack by Inferno
			for i = 0, 8
				local fa = (i*(ANGLE_180/4))
				local dust = P_SpawnMobj(mo.x, mo.y, mo.z, MT_SPINDUST)
				dust.momx = FixedMul(sin(fa),20*FRACUNIT)
				dust.momy = FixedMul(cos(fa),20*FRACUNIT)
			end

			local GBOUNCE_HEIGHT = BASE_GBOUNCE_HEIGHT
			if p.powers[pw_super]
				GBOUNCE_HEIGHT = 6*$/5 --Increase modifier by 20 percent when Super
			end
			if mo.eflags & MFE_UNDERWATER
				GBOUNCE_HEIGHT = 3*$/5 --Decrease modifier by 40 percent when underwater
			end
			P_SetObjectMomZ(mo, GBOUNCE_HEIGHT, false)
			p.ghostTrail = 14
			S_StartSound(mo, sfx_mspogo)
		end
	end
	

	if p.chrgWBounce > 0
		p.chrgWBounce = $-1
	end

	--Perform sprite scaling and keep Z-momentum constant while stomping
	if p.chrgStomping
		P_SetObjectMomZ(mo, p.STOMPSPEED, false)
		if P_PlayerInPain(p)
		or mo.eflags & MFE_GOOWATER
		or mo.state ~= S_PLAY_ROLL
			p.chrgStomping = false
		end

		if mo.spriteyscale < 3*FRACUNIT/2
			mo.spriteyscale = $+FRACUNIT/10
		end
		if mo.spritexscale > 9*FRACUNIT/10
			mo.spritexscale = $-FRACUNIT/10
		end

	elseif p.prevChrgStomp
		mo.spritexscale = FRACUNIT
		mo.spriteyscale = FRACUNIT
		p.prevChrgStomp = false
	end
end)

//Behaviour for when an enemy is damaged
addHook("MobjDamage", function(t, i, s)
	if not (i and i.player and i.skin == SKIN) return end
	local p = i.player
	if not P_IsObjectOnGround(i)
	and t.valid
		if p.chrgStomping
			p.chrgStomping = false
			p.prevChrgStomp = true
			i.momz = $/2
		end
		if t.flags & MF_ENEMY|MF_BOSS

			--Recover a single charge if the player has less than one
			if p.charges < 1
				p.charges = 1
			end
			ResetAbilities(p)
		end
	end
end)

//Bustable FOF detection code from Krabs's Rebound Dash
addHook("MobjLineCollide", function(mo, line)
	if not mo or not mo.valid return end
	if not (mo.skin == SKIN) return end
	local p = mo.player
	if (p == nil) or (p.playerstate != PST_LIVE) return end

	local side = P_PointOnLineSide(mo.x, mo.y, line)
	local sector = nil

	if side == 1
		sector = line.frontsector
	else
		sector = line.backsector
	end

	if sector == nil return end

	for rover in sector.ffloors()
		if not (rover.flags & FF_EXISTS) or not (rover.flags & FF_BLOCKPLAYER)
			continue
		end

		local topheight = rover.topheight
		local bottomheight = rover.bottomheight

		if (rover.t_slope)
			topheight = P_GetZAt(rover.t_slope, mo.x, mo.y)
		end
		if (rover.b_slope)
			bottomheight = P_GetZAt(rover.b_slope, mo.x, mo.y)
		end

		if (topheight < mo.z)
			continue
		end

		if (bottomheight > mo.z + mo.height)
			continue
		end

		if (rover.flags & FF_BUSTUP) and not (rover.flags & FF_STRONGBUST)
			p.bustsector = sector
			p.bustrover = rover
		end
		break
	end
end)

//Behaviour when movement is blocked
addHook("MobjMoveBlocked", function(mo)
	if not (mo and mo.health and mo.valid and mo.skin == SKIN) return end
	local p = mo.player

	--Bounce when hitting a wall with a fully charged Charge Thok
	if p.chrgWBounce
		P_BounceMove(mo)

		--Set player momentum
		mo.momx = 4*$/11
		mo.momy = 4*$/11
		P_SetObjectMomZ(mo, 12*FRACUNIT, false)

		S_StartSound(mo, sfx_mspogo)
		p.chrgWBounce = 0
		p.chrgSparkle = false
		p.ghostTrail = 14
		mo.state = S_PLAY_POSTWBNC
		p.pflags = $|PF_NOJUMPDAMAGE

		--Dust from Krabs's Rebound Dash
		for i = 0, 9
			local dust = P_SpawnMobjFromMobj(mo, 0, 0, mo.scale*18, MT_SPINDUST)
			dust.momx = (mo.momx*2)+(P_RandomRange(-10, 10)*FRACUNIT)
			dust.momy = (mo.momy*2)+(P_RandomRange(-10, 10)*FRACUNIT)
			dust.momz = P_RandomRange(-3, 3)*FRACUNIT
			dust.scale = P_RandomRange(50, 125)*FRACUNIT/100
			dust.destscale = 0
			dust.state = S_BOSSEXPLODE
		end

		--Break bustable FOFs from Krabs's Rebound Dash
		if p.bustsector and p.bustrover
			EV_CrumbleChain(p.bustsector, p.bustrover)
			p.bustsector = nil
			p.bustrover = nil

			--Recover a single charge if the player has less than one
			if p.charges < 1
				p.charges = 1
			end
			ResetAbilities(p)
		end
	end
end, MT_PLAYER)