local SKIN = "jetmetalsonic"

//Set default values
local SEARCH_DIST = 6*RING_DIST/5	--Search distance for valid homing attack targets
local SEARCH_ANGLE = ANGLE_45		--Angle within which to search for valid homing targets

local SPEED_MOD = 2*FRACUNIT/3		--Modifier for movement speed during homing attack
local HOME_TIMER = 3*TICRATE		--Homing attack duration

//Code re-used and modified from CobaltBW's Fang Popgun Modifications
local function JetLookForEnemies(p, ringdist, span, above, disregard) 
	local twod = (twodlevel or p.mo.flags2 & MF2_TWOD)
	if (twod)
		ringdist = $/2
		span = 2*$/3
	end

	local maxdist = FixedMul(ringdist, p.mo.scale)
	local closestdist = 0
	local closestmo = nil
	local nonenemiesdisregard = disregard
	searchBlockmap("objects",function(pmo,mo)
		if (mo.flags & MF_NOCLIPTHING) return end
		if (mo.health <= 0) return end

		if not(mo.p)
		and (!((mo.flags & (MF_ENEMY|MF_BOSS|MF_MONITOR)
		and (mo.flags & MF_SHOOTABLE)) or (mo.flags & MF_SPRING)) == !(mo.flags2 & MF2_INVERTAIMABLE))
			return
		end
		if (mo == pmo) return end
		if (mo.flags2 & MF2_FRET) return end
		if (mo.flags & nonenemiesdisregard) return end

		//Disallow targeting teammates
		if mo.p and (mo.p.spectator or gametyperules&GTR_FRIENDLY or G_GametypeHasTeams() and mo.p.ctfteam == p.ctfteam)
			return
		end

		//Do angle/distance checks
		local zdist = (pmo.z + pmo.height/2) - (mo.z + mo.height/2)
		local dist = P_AproxDistance(pmo.x - mo.x, pmo.y - mo.y)
		//CBW:	Made the angle checks their own locals, for readability purposes
		//		I also unsigned the angle checks, which appears to correct failed OutOfBounds checks above the player.
		local xyz_angle = abs(R_PointToAngle2(0, 0, dist, zdist))
		local xy_angle = abs(R_PointToAngle2(
				pmo.x + P_ReturnThrustX(pmo, pmo.angle, pmo.radius),
				pmo.y + P_ReturnThrustY(pmo, pmo.angle, pmo.radius),
				mo.x, mo.y) - pmo.angle)

		dist = P_AproxDistance(dist, zdist)
		if (dist > maxdist)
			return -- out of range
		end
		if (xyz_angle > span)
			return -- Don't home outside of desired angle!
		end

		if not above -- Don't home above the player if not desired!
			if pmo.eflags & MFE_VERTICALFLIP
			and zdist > 0
				return
			elseif zdist < 0
				return
			end
		end

		if (twod
		and abs(pmo.y-mo.y) > pmo.radius)
			return -- not in your 2d plane
		end

		if ((closestmo and closestmo.valid) and (dist > closestdist))
			return
		end
		if (xy_angle > span)
			return -- behind back
		end

		if not (P_CheckSight(pmo, mo))
			return -- out of sight
		end

		closestmo = mo
		closestdist = dist
	end,p.mo,p.mo.x-maxdist,p.mo.x+maxdist,p.mo.y-maxdist,p.mo.y+maxdist)
	return closestmo
end

local function JetHoming(p, t, spdmod)
	local mo = p.mo
	mo.angle = R_PointToAngle2(mo.x, mo.y, t.x, t.y)

	local dist = P_AproxDistance(P_AproxDistance(t.x - mo.x, t.y - mo.y), t.z - mo.z)
	local speed =  FixedMul(FixedMul(p.actionspd, spdmod), mo.scale)
	if mo.eflags & MFE_UNDERWATER
		speed = 3*$/5 --Reduce speed by 40% when underwater
	end

	mo.momx = FixedMul(FixedDiv(t.x - mo.x, dist), speed)
	mo.momy = FixedMul(FixedDiv(t.y - mo.y, dist), speed)
	mo.momz = FixedMul(FixedDiv(t.z - mo.z, dist), speed)
end

addHook("PlayerThink", function(p)
	local mo = p.mo
	if not (mo and mo.valid and mo.health and mo.skin == SKIN) return end

	if (maptol & TOL_NIGHTS) return end
	if p.powers[pw_carry] return end

	if not p.Homing
	and not (p.pflags & PF_THOKKED)
	and not P_IsObjectOnGround(mo)
	and not P_PlayerInPain(p)
		mo.tracer = JetLookForEnemies(p, SEARCH_DIST, SEARCH_ANGLE, true, MF_SPRING|MF_MISSILE|MF_MONITOR) --Search for valid targets
		if mo.tracer and mo.tracer.valid and mo.tracer.health
		and not (mo.state == S_PLAY_SPRING or mo.state == S_PLAY_FALL)
		and not p.charging
			P_SpawnLockOn(p, mo.tracer, S_LOCKON2) --Spawn LockOn object
		end
	end

	--Perform homing attack
	if P_PlayerInPain(p)
	or (mo.eflags & MFE_SPRUNG)
		p.Homing = 0
	end
	if p.Homing
		mo.state = S_PLAY_DASH
		p.pflags = ($ & ~PF_NOJUMPDAMAGE)|PF_STASIS
		JetHoming(p, mo.tracer, SPEED_MOD)
		p.Homing = $-1

		if not p.Homing
			mo.state = S_PLAY_FALL
			p.pflags = $|PF_NOJUMPDAMAGE|PF_THOKKED
		end
	end
end)

//Hook behaviour to SPIN while jumping
addHook("JumpSpinSpecial", function(p)
	local mo = p.mo
	if not (mo and mo.health and mo.valid and mo.skin == SKIN) return end

	--Go super, checks adapted and modified from Reveries by Katmint
	if All7Emeralds(emeralds) and p.rings > 49 and p.playerstate == PST_LIVE and p.charflags & SF_SUPER
	and not p.powers[pw_super] and not p.powers[pw_carry] and not p.powers[pw_shield]
	and not (p.pflags & (PF_SPINNING|PF_THOKKED|PF_SPINDOWN))
	and not (maptol & TOL_NIGHTS)
	and not p.panim == PA_ABILITY and not p.panim == PA_ABILITY2 and not p.panim == PA_PAIN and not p.panim == PA_ETC
	and not p.mo.state == S_PLAY_CHRG and not p.chrgStomping
	and not p.Homing and not p.charging
	and not P_IsObjectOnGround(p.mo)
		P_DoSuperTransformation(p)
		return true

	--Perform the Homing Thok
	elseif not p.Homing
	and not (p.pflags & (PF_SPINDOWN|PF_THOKKED))
	and mo.tracer and mo.tracer.valid and mo.tracer.health
		S_StartSound(mo, sfx_beflap)
		mo.angle = R_PointToAngle2(mo.x, mo.y, mo.tracer.x, mo.tracer.y)
		SpawnLaunchEffect(p, 1, 3*FRACUNIT/5, p.effcolor)
		p.Homing = HOME_TIMER
		p.pflags = $ & ~PF_NOJUMPDAMAGE
		p.chrgStomping = false
		p.ghostTrail = 0
		return true
	end
end)

//Homing Thok has priority over shield abilities
/*addHook("ShieldSpecial", function(p)
	if not (p.mo and p.mo.health and p.mo.valid and p.mo.skin == SKIN) return end
	if (p.mo.tracer and p.mo.tracer.valid) or p.Homing or p.pflags & PF_SPINDOWN return true end
end)*/

//Behaviour for when an enemy is damaged
addHook("MobjDamage", function(t, i, s)
	if not (i and i.player and i.skin == SKIN) return end
	local p = i.player
	if not P_IsObjectOnGround(i)
	and t.valid
		if (t.flags & MF_ENEMY|MF_BOSS|MF_MONITOR)
		and p.Homing
			local speed = p.actionspd
			if i.eflags & MFE_UNDERWATER
				speed = 3*$/5 --Reduce speed by 40% when underwater
			end
			P_InstaThrust(i, i.angle, FixedMul(speed/6, i.scale))
			P_SetObjectMomZ(i, speed/10, false)
			p.pflags = $ & ~PF_THOKKED
			i.state = S_PLAY_POSTHOME
			p.Homing = 0
		end
	end
end)