local SKIN = "jetmetalsonic"

local function SpawnSparkEffect(mo, color, freq, tics)
	if leveltime % freq ~= 0 return end
	local x_adjust = FixedMul(FixedMul(P_RandomRange(-FRACUNIT/2, FRACUNIT/2 - 1), mo.scale), mo.radius)
	local y_adjust = FixedMul(FixedMul(P_RandomRange(-FRACUNIT/2, FRACUNIT/2 - 1), mo.scale), mo.radius)
	local z_adjust = FixedMul(FixedMul(P_RandomFixed(), mo.scale), mo.height)
	if mo.eflags & MFE_VERTICALFLIP
		z_adjust = mo.height - mo.height / 3 - FixedMul(FixedMul(P_RandomFixed(), mo.scale), mo.height)
	end

	local spark = P_SpawnMobj(mo.x + x_adjust, mo.y + y_adjust, mo.z + z_adjust, MT_WATERZAP)
	spark.tics = tics
	spark.color = color
	spark.colorized = true
	spark.blendmode = AST_ADD
	if mo.eflags & MFE_VERTICALFLIP
		spark.eflags = $|MFE_VERTICALFLIP
	end

	if not mo.player return end
	if mo.player.powers[pw_super]
		P_InstaThrust(spark, R_PointToAngle2(mo.x, mo.y, spark.x, spark.y), FixedMul(3*FRACUNIT, mo.scale))
		P_SetObjectMomZ(spark, 7*FRACUNIT/2, false)
	end
end

local function BoosterJetFume(mo, ghost)
	local x_adjust = P_ReturnThrustX(mo, mo.angle, -17*mo.scale)
	local y_adjust = P_ReturnThrustY(mo, mo.angle, -17*mo.scale)
	if not (mo.eflags & MFE_UNDERWATER)
		local jetfume = P_SpawnMobjFromMobj(mo, x_adjust, y_adjust, 0, MT_THOK)
		jetfume.fuse = 1
		jetfume.sprite = SPR_JETF
		jetfume.colorized = true
		jetfume.color = mo.tracer.player.effcolor
		jetfume.scale = mo.scale/2

		if leveltime % 3 == 0
			jetfume.frame = A|FF_FULLBRIGHT
		elseif leveltime % 3 == 1
			jetfume.frame = B|FF_FULLBRIGHT
		else
			jetfume.frame = C|FF_FULLBRIGHT
		end

		if mo.eflags & MFE_VERTICALFLIP
			jetfume.eflags = $|MFE_VERTICALFLIP
			P_TeleportMove(jetfume, jetfume.x, jetfume.y, jetfume.z + FixedMul(32*FRACUNIT, mo.scale))
		end

		if ghost
			P_SpawnGhostMobj(jetfume)
		end

	else
		local bubble = P_SpawnMobjFromMobj(mo, 0, 0, 0, MT_SMALLBUBBLE)
		if mo.eflags & MFE_VERTICALFLIP
			P_TeleportMove(bubble, mo.x + x_adjust, mo.y + y_adjust, mo.z + mo.scale)
			bubble.eflags = $|MFE_VERTICALFLIP
		else
			P_TeleportMove(bubble, mo.x + x_adjust, mo.y + y_adjust, mo.z - mo.scale)
		end
	end
end

//Reset stuff when the Player spawns
addHook("PlayerSpawn", function(p)
	if not (p.mo and p.mo.health and p.mo.valid and p.mo.skin == "jetmetalsonic") return end
	ResetAbilities(p)
	ResetCharge(p)
	p.boostersalreadyexploded = false
	p.combocount = 0
end)

//Recurl by pressing JUMP or SPIN while in Spring state
addHook("SpinSpecial", function(p)
	local mo = p.mo
	if not (mo and mo.valid and mo.health and mo.skin == SKIN) return end
	if mo.state == S_PLAY_SPRING
	and not (p.pflags & PF_SPINDOWN)
		p.pflags = $|PF_JUMPED|PF_SPINDOWN
		mo.state = S_PLAY_ROLL
		mo.eflags = $ & ~MFE_SPRUNG
		S_StartSound(mo, sfx_3db16)
		return true
	end
end)
addHook("JumpSpecial", function(p)
	local mo = p.mo
	if not (mo and mo.valid and mo.skin == SKIN) return end
	if mo.state == S_PLAY_SPRING
	and not (p.pflags & PF_JUMPDOWN)
		p.pflags = $|PF_JUMPED|PF_JUMPDOWN
		mo.state = S_PLAY_ROLL
		mo.eflags = $ & ~MFE_SPRUNG
		S_StartSound(mo, sfx_3db16)
	end
end)

//Jet boosters and aura hooked into ThinkFrames
addHook("ThinkFrame", do
	for p in players.iterate
		local mo = p.mo
		if mo and mo.valid and mo.skin == SKIN

			--Spawn boosters if they do not exist yet
			if not p.jetmetbstrA or not p.jetmetbstrA.valid
				p.jetmetbstrA = P_SpawnMobj(mo.x, mo.y, mo.z, MT_JETMETBSTERA)
				local mo_A = p.jetmetbstrA
				mo_A.tracer = mo
				mo_A.tracecolor = true
				mo_A.skin = mo.skin
				mo_A.sprite = SPR_PLAY
				mo_A.sprite2 = SPR2_JMSB
			end
			if not p.jetmetbstrB or not p.jetmetbstrB.valid
				p.jetmetbstrB = P_SpawnMobj(mo.x, mo.y, mo.z, MT_JETMETBSTERB)
				local mo_B = p.jetmetbstrB
				mo_B.tracer = mo
				mo_B.tracecolor = true
				mo_B.skin = mo.skin
				mo_B.sprite = SPR_PLAY
				mo_B.sprite2 = SPR2_JMSB
			end

			local mo_A = p.jetmetbstrA
			local mo_B = p.jetmetbstrB

			if mo_A and mo_A.valid
			and mo_B and mo_B.valid
				local lastposx_A = mo_A.x
				local lastposy_A = mo_A.y
				local lastposz_A = mo_A.z
				local lastposx_B = mo_B.x
				local lastposy_B = mo_B.y
				local lastposz_B = mo_B.z
				local vertflip = false
				if mo.eflags & MFE_VERTICALFLIP
					vertflip = true
				end
				mo_A.scale = mo.scale
				mo_B.scale = mo.scale
				mo_A.colorized = mo.colorized
				mo_B.colorized = mo.colorized

				--Have the boosters bob up and down
				if leveltime % (2*TICRATE) < TICRATE
					p.relative_z = 40<<FRACBITS + (leveltime % TICRATE) / 8<<FRACBITS
				else
					p.relative_z = 40<<FRACBITS + 34<<FRACBITS/8 - (leveltime % TICRATE) / 8<<FRACBITS
				end
				A_CapeChase(mo_A, p.relative_z + 1, -10<<FRACBITS + 20)
				A_CapeChase(mo_B, p.relative_z + 1, -10<<FRACBITS - 20)

				--Spawn an explosion if the boosters cease to exist
				if not (mo_A and mo_A.valid and mo_B and mo_B.valid)
					if p.boostersalreadyexploded == true return end

					local explosionA = P_SpawnMobj(lastposx_A, lastposy_A, lastposz_A, MT_THOK)
					explosionA.state = S_SONIC3KBOSSEXPLOSION1
					explosionA.renderflags = $|RF_FULLBRIGHT
					if vertflip == true
						explosionA.eflags = $|MFE_VERTICALFLIP
					end

					local explosionB = P_SpawnMobj(lastposx_B, lastposy_B, lastposz_B, MT_THOK)
					explosionB.state = S_SONIC3KBOSSEXPLOSION1
					explosionB.renderflags = $|RF_FULLBRIGHT
					if vertflip == true
						explosionB.eflags = $|MFE_VERTICALFLIP
					end

					p.boostersalreadyexploded = true
					return
				end

				--Boosters crackle with electricity when super
				if p.powers[pw_super]
					SpawnSparkEffect(mo_A, p.effcolor, 3, 4)
					SpawnSparkEffect(mo_B, p.effcolor, 3, 4)
				end
			end

			--Spawn jetfumes under specific circumstances
			if (mo.state == S_PLAY_CHRGJMP or mo.state == S_PLAY_CHRGTHK or p.chrgStomping)
			and p.ghostTrail
				if not p.chrgStomping
					local x_adjust = P_ReturnThrustX(mo, mo.angle, FixedMul(-15*FRACUNIT, mo.scale))
					local y_adjust = P_ReturnThrustY(mo, mo.angle, FixedMul(-15*FRACUNIT, mo.scale))
					local z_adjust = FixedMul(16*FRACUNIT, mo.scale)
					if mo.eflags & MFE_VERTICALFLIP
						z_adjust = FixedMul(-16*FRACUNIT, mo.scale)
					end
					if mo.state == S_PLAY_CHRGJMP
						z_adjust = 0
					end

					if not (mo.eflags & MFE_UNDERWATER)
						local jetfume = P_SpawnMobj(mo.x + x_adjust, mo.y + y_adjust, mo.z + z_adjust, MT_THOK)
						jetfume.fuse = 1
						jetfume.colorized = true
						jetfume.color = p.effcolor
						jetfume.sprite = SPR_JETF
						jetfume.scale = 2*mo.scale/3
						if mo.eflags & MFE_VERTICALFLIP
							jetfume.eflags = $|MFE_VERTICALFLIP
						end

						if leveltime % 3 == 0
							jetfume.frame = A|FF_FULLBRIGHT
						elseif leveltime % 3 == 1
							jetfume.frame = B|FF_FULLBRIGHT
						else
							jetfume.frame = C|FF_FULLBRIGHT
						end
						P_SpawnGhostMobj(jetfume)

					--if underwater, spawn bubbles instead
					else
						P_SpawnMobj(mo.x+x_adjust, mo.y+y_adjust, mo.z+z_adjust, MT_SMALLBUBBLE)
					end
				end
				BoosterJetFume(p.jetmetbstrA, true)
				BoosterJetFume(p.jetmetbstrB, true)
			end
			if mo.state == S_PLAY_RUN
			or mo.state == S_PLAY_DASH
				if p.Homing
					BoosterJetFume(p.jetmetbstrA, true)
					BoosterJetFume(p.jetmetbstrB, true)
					p.followmobj.color = p.effcolor
					P_SpawnGhostMobj(p.followmobj)
				else
					BoosterJetFume(p.jetmetbstrA, false)
					BoosterJetFume(p.jetmetbstrB, false)
				end
			end

			--Use aura effect under certain circumstances
			if p.Homing
			or p.chrgStomping
			or mo.state == S_PLAY_CHRGJMP
			or mo.state == S_PLAY_CHRGTHK
				local aura = P_SpawnMobj(mo.x, mo.y, mo.z, MT_THOK)
				aura.fuse = 1
				aura.angle = mo.angle
				aura.sprite = SPR_JMSA
				aura.color = p.effcolor
				aura.dispoffset = 1
				if not (aura.flags2 & MF2_LINKDRAW)
					aura.flags2 = $|MF2_LINKDRAW
				end
				if mo.eflags & MFE_VERTICALFLIP
					aura.eflags = $|MFE_VERTICALFLIP
					P_TeleportMove(aura, aura.x, aura.y, aura.z - FixedMul(12*FRACUNIT, mo.scale))
				end

				if p.Homing
				or mo.state == S_PLAY_CHRGTHK
					aura.frame = A
				elseif p.chrgStomping
					aura.frame = B
				else
					aura.frame = C
				end

				if leveltime % 2 == 1
					aura.frame = $|FF_TRANS50|FF_FULLBRIGHT
				else
					aura.frame = $|FF_TRANS70|FF_FULLBRIGHT
				end
			end
		end
	end
end)

//Miscellaneous PlayerThinker stuff
addHook("PlayerThink", function(p)
	local mo = p.mo
	if not (mo and mo.valid and mo.health and mo.skin == SKIN) return end

	SpawnSparkEffect(mo, p.effcolor, 1, P_RandomRange(4, 8))
	if p.powers[pw_super]
		SpawnSparkEffect(mo, p.effcolor, 3, 8)
	end

	if (maptol & TOL_NIGHTS) return end
	if p.powers[pw_carry] return end

	--Do not air walk or air stand
	if p.pflags & ~(PF_JUMPED|PF_SPINNING)
	and not P_PlayerInPain(p)
	and not P_IsObjectOnGround(mo)
	and not p.powers[pw_nocontrol]
	and not p.charging
		if mo.state == S_PLAY_STND
		or mo.state == S_PLAY_WAIT
		or mo.state == S_PLAY_WALK
		or mo.state == S_PLAY_RUN
		or (mo.state == S_PLAY_DASH and not p.Homing)
			mo.state = S_PLAY_SPRING
		end
	end

	--Create trail when required
	if not p.ghostTrail
		p.ghostTrail = 0
	end
	if p.ghostTrail > 0
	or (P_IsObjectOnGround(mo) and p.pflags & PF_SPINNING and not (p.pflags & PF_STARTDASH))
		local playerghost = P_SpawnGhostMobj(mo)
		local boosterghostA = P_SpawnGhostMobj(p.jetmetbstrA)
		local boosterghostB = P_SpawnGhostMobj(p.jetmetbstrB)
		playerghost.colorized = true
		playerghost.color = p.effcolor
		boosterghostA.colorized = true
		boosterghostA.color = p.effcolor
		boosterghostB.colorized = true
		boosterghostB.color = p.effcolor
		if p.ghostTrail > 0
			p.ghostTrail = $-1
		end
	end

	--Use Dashing animation when running fast enough
	if P_IsObjectOnGround(mo)
	and mo.state == S_PLAY_RUN
	and FixedHypot(p.rmomx, p.rmomy) > 32*FRACUNIT
		mo.state = S_PLAY_DASH
	elseif P_IsObjectOnGround(mo)
	and mo.state == S_PLAY_DASH
	and FixedHypot(p.rmomx, p.rmomy) < 32*FRACUNIT
		mo.state = S_PLAY_RUN
	end
end)

//Set followmobj color in PostThinkFrame
addHook("PostThinkFrame", do
	for p in players.iterate
		if not (p.mo and p.mo.health and p.mo.valid and p.mo.skin == SKIN) return end
		p.followmobj.colorized = true
		p.followmobj.color = p.effcolor
	end
end)

//Hook custom object properties into the generic MobjThinker
addHook("MobjThinker", function(mo)
	if mo.tracecolor
		mo.color = mo.tracer.color
	end

	if mo.ischrgeff
		A_CapeChase(mo, 48<<FRACBITS+1, -40<<FRACBITS)
		if mo.valid
			mo.angle = mo.tracer.angle + ANGLE_90
		end
	end
end)